#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for the Router """

import json
import logging
import sys
from typing import Dict
from unittest.mock import MagicMock, patch

import pytest
from _pytest.capture import CaptureFixture
from messaging.router import Router, on_message

# pylint: disable=E0401, E1101, W0621


pytest_plugins = ["testing.utils.conftest"]

logger = logging.getLogger("router")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class CallbackTester:
    """You won't be believe it, but this thing tests message callback."""

    last_message = None

    @on_message(type="workflow-request-complete")
    def callback(self, **message):
        self.last_message = {"callback": "callback", "message": message}

    @on_message(type="workflow-request-ready", subject={"workflow_name": "test"})
    def callback_nested(self, **message):
        # print(f"CallbackTester.callback_workflow_ready> WORKFLOW READY! {message}")
        self.last_message = {"callback": "callback_nested", "message": message}

    @on_message(type="type1")
    @on_message(type="type2")
    def callback_two_decorators(self, **message: Dict):
        """
        Sets last message to a callback with two, count 'em, TWO (2)! decorators

        :param message:
        :return:
        """
        self.last_message = {"callback": "callback_two_decorators", "message": message}


@pytest.fixture
def callback_tester(test_router: Router):
    """
    A callback tester for our tests.

    :param test_router: our test Router
    :return:
    """
    tester = CallbackTester()
    test_router.register(tester)
    return tester


@patch("messaging.router.Thread")
@patch("messaging.router.Messenger")
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_create_messenger(mock_messenger: MagicMock, mock_thread: MagicMock):
    """
    Tests that Router correctly creates a messenger and a thread for said messenger

    :return:
    """
    test_router = Router("test_create_messenger")
    assert test_router.messenger
    assert test_router.messenger_thread
    assert test_router.messenger_thread.daemon is True
    assert test_router.messenger_thread.start.call_count == 1


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_register_callback(test_router: Router, callback_tester: CallbackTester):
    """
    Tests that Router callback registration using the on_message decorator
    works with workflow_complete message

    :param test_router: our router
    :param callback_tester: the tester

    :return:
    """
    message_type = "workflow-request-complete"
    test_router.send_message(
        application="archive_events",
        service="capability",
        subject={
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "workflow_name": "null",
            "argument": "-g",
            "state": "Complete",
        },
        type=message_type,
    )
    assert callback_tester.last_message is not None
    assert callback_tester.last_message["callback"] == "callback"

    last_message = callback_tester.last_message["message"]
    logger.debug(f">>> LAST MESSAGE: {last_message}")
    assert last_message["service"] == "capability"
    assert last_message["type"] == message_type
    assert last_message["subject"] == {
        "type": "WorkflowRequest",
        "workflow_request_id": -1,
        "workflow_name": "null",
        "argument": "-g",
        "state": "Complete",
    }


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_register_callback_nested(test_router: Router, callback_tester: CallbackTester):
    """
    Tests that Router callback registration using the on_message decorator
    works with nested dictionary entries

    :param test_router: our router
    :param callback_tester: our tester

    :return:
    """
    message_type = "workflow-request-ready"
    test_router.send_message(
        application="mr_books",
        service="capability",
        subject={
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "workflow_name": "test",
            "argument": "-g",
            "state": "Ready",
        },
        type=message_type,
    )
    assert callback_tester.last_message is not None
    assert callback_tester.last_message["callback"] == "callback_nested"

    last_message = callback_tester.last_message["message"]
    assert last_message["type"] == message_type


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_register_callback_two_types(test_router: Router, callback_tester: CallbackTester):
    """
    Tests that Router callback registration using the on_message decorator
    works with nested dictionary entries

    :param test_router: the Router to test
    :param callback_tester: the CallbackTester we'll use

    :return:
    """
    # Message #1
    test_router.send_message(
        application="some_incredibly_important_event",
        service="capability",
        subject={
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "workflow_name": "test",
            "argument": "-g",
            "state": "Ready",
        },
        type="type1",
    )
    assert callback_tester.last_message is not None
    assert callback_tester.last_message["callback"] == "callback_two_decorators"

    last_message = callback_tester.last_message["message"]
    assert last_message["type"] == "type1"

    # Message #2
    test_router.send_message(
        application="I love sending messages, don't you?",
        service="capability",
        subject={
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "workflow_name": "test",
            "argument": "-g",
            "state": "Ready",
        },
        type="type2",
    )
    assert callback_tester.last_message is not None
    assert callback_tester.last_message["callback"] == "callback_two_decorators"

    last_message = callback_tester.last_message["message"]
    assert last_message["type"] == "type2"


def fake_callback(**message: Dict):
    """
    Pretend we're doing a callback.

    :param message:
    :return:
    """
    print(message["type"])


def test_receive_message(test_router: Router, capsys: CaptureFixture):
    """
    Tests that the correct message-receiving procedure is undergone
    """
    test_router.callbacks = {'{"type": "hello-world"}': [fake_callback]}
    test_router.receive_message(
        json.dumps(
            dict(
                service="capability",
                application="this is your message. this is your message on drugs.",
                subject={
                    "type": "WorkflowRequest",
                    "workflow_request_id": -1,
                    "workflow_name": "test",
                    "argument": "-g",
                    "state": "Ready",
                },
                type="hello-world",
            )
        )
    )
    assert capsys.readouterr().out == "hello-world\n"


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_send_message(test_router: Router):
    """
    Tests that router correctly routes message to its messenger
    """
    test_router.messenger = MagicMock()
    test_router.send_message(hello="world")
    assert test_router.messenger.send_message.call_count == 1


def test_matches():
    """
    Tests that Router.matches will correctly match patterns with messages
    """
    message = {
        "type": "capability_event",
        "subject": {"type": "capability_request"},
        "status": "complete",
        "capability_request": {
            "id": 12,
            "capability_name": "test_download",
            "state": "complete",
        },
    }

    # positive matches
    assert Router.matches({"type": "capability_event"}, message)
    assert Router.matches({"subject": {"type": "capability_request"}}, message)
    assert Router.matches({"type": "capability_event", "status": "complete"}, message)
    assert Router.matches({"capability_request": {"capability_name": "test_download"}}, message)
    assert Router.matches({}, message)

    # negative matches
    assert not Router.matches({"type": "workflow_event"}, message)
    assert not Router.matches({"type": "capability_event", "status": "started"}, message)
    assert not Router.matches({"capability_request": {"capability_name": "not_test_download"}}, message)
