#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Optional, Set
from unittest.mock import MagicMock

import pytest
from hypothesis import given
from hypothesis import strategies as st

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
    CapabilityVersionFile,
)
from workspaces.capability.services.capability_info import CapabilityInfo

pytest_plugins = ["testing.utils.conftest"]
SQLITE_MIN_INT = -9223372036854775808
SQLITE_MAX_INT = 9223372036854775807
SAMPLE_CAPABILITY_NAMES = ["test1", "test2", "test3"]

# Register CapabilityRequest JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityRequest))
# >>> def test(generated_request_json: Dict):
# >>>   request = CapabilityRequest.from_json(generated_request_json)
st.register_type_strategy(
    CapabilityRequest,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityRequest"),
            "id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "capability_name": st.sampled_from(SAMPLE_CAPABILITY_NAMES),
            "state": st.sampled_from([name for name, _ in CapabilityRequestState.__members__.items()]),
            "parameters": st.one_of(st.lists(st.text()), st.none()),
            "ingested": st.booleans(),
            "created_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "updated_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "current_execution": st.none(),
        }
    ),
)

# Register CapabilityVersion JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityVersion))
# >>> def test(generated_version_json: Dict):
# >>>   version = CapabilityVersion.from_json(generated_version_json)
st.register_type_strategy(
    CapabilityVersion,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityVersion"),
            "capability_request_id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "version_number": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "parameters": st.one_of(st.lists(st.text()), st.none()),
            "workflow_metadata": st.one_of(
                st.none(), st.dictionaries(st.text(min_size=1), st.text(min_size=1), min_size=1)
            ),
            "files": st.none(),
        }
    ),
)

# Register CapabilityVersionFile JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityVersionFile))
# >>> def test(generated_version_file_json: Dict):
# >>>   version_file = CapabilityVersionFile.from_json(generated_version_file_json)
st.register_type_strategy(
    CapabilityVersionFile,
    st.fixed_dictionaries({"filename": st.text(min_size=1), "content": st.binary(min_size=10).map(bytes.hex)}),
)

# Register CapabilityExecution JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityExecution))
# >>> def test(generated_execution_json: Dict):
# >>>   execution = CapabilityExecution.from_json(generated_request_json)
st.register_type_strategy(
    CapabilityExecution,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityExecution"),
            "id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "state_name": st.sampled_from(["Start", "Queued", "Executing", "Foo", "Bar"]),
            "queue_state": st.sampled_from([name for name, _ in QueueState.__members__.items()]),
            "capability_request_id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "version_number": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "current_workflow_request_id": st.just(-11),
            "delivery_url": st.from_regex(r"https:\/\/dl-nrao-unittest\.aoc\.nrao\.edu(?:\/[A-z0-9]*)+"),
            "created_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "updated_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
        }
    ),
)


@given(st.text(), st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT), st.booleans())
def test_create_capability(mock_capability_info: CapabilityInfo, name: str, max_jobs: int, has_image_products: bool):
    capability = mock_capability_info.create_capability(name, max_jobs, has_image_products)
    mock_capability_info.save_entity.assert_called_with(capability)


@given(
    st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
    st.booleans(),
    st.booleans(),
)
def test_edit_capability(mock_capability_info: CapabilityInfo, max_jobs: int, enabled: bool, has_img_prdcts: bool):
    """
    Tests that edit_capability method properly follows its logic
    """

    cap_name = "null"

    # Test edit with all inputs
    assert mock_capability_info.edit_capability(cap_name, max_jobs, enabled, has_img_prdcts)
    # Test edit with just name, max_jobs, and enabled
    assert mock_capability_info.edit_capability(cap_name, max_jobs=max_jobs, enabled=enabled)
    # Test edit with just name and max_jobs
    assert mock_capability_info.edit_capability(cap_name, max_jobs=max_jobs)
    # Test edit with just name
    assert mock_capability_info.edit_capability(cap_name)
    # Test "unable to edit" condition
    orig_session = mock_capability_info.session
    mock_capability_info.session = MagicMock()
    mock_capability_info.session.update.return_value = 0
    assert not mock_capability_info.edit_capability(cap_name, max_jobs=max_jobs)
    mock_capability_info.session = orig_session

    capability = mock_capability_info.lookup_capability(cap_name)
    assert capability.max_jobs == max_jobs
    assert capability.enabled is enabled
    assert capability.has_image_products is has_img_prdcts


def test_pause_capability(mock_capability_info: CapabilityInfo):
    cap_name = "null"
    capability = mock_capability_info.lookup_capability(cap_name)
    assert capability.paused is False
    mock_capability_info.pause_capability(cap_name)
    assert capability.paused is True


def test_unpause_capability(mock_capability_info: CapabilityInfo):
    cap_name = "null"
    capability = mock_capability_info.lookup_capability(cap_name)
    assert capability.paused is True
    mock_capability_info.unpause_capability(cap_name)
    assert capability.paused is False


@given(st.text(), st.one_of(st.lists(st.text()), st.none()))
@pytest.mark.usefixtures("mock_capability_requests")
def test_create_capability_request(
    mock_capability_info: CapabilityInfo,
    cap_name: str,
    parameters: Optional[List[str]],
):
    capability_request = mock_capability_info.create_capability_request(cap_name, parameters)
    mock_capability_info.save_entity.assert_called_with(capability_request)


@given(st.one_of(st.lists(st.text()), st.none()))
def test_create_capability_version(
    mock_capability_info: CapabilityInfo,
    parameters: Optional[List[str]],
):
    req_id = -1
    version = mock_capability_info.create_capability_version(req_id, parameters)
    mock_capability_info.save_entity.assert_called_with(version)


@pytest.mark.usefixtures("mock_capability_requests")
def test_create_execution(mock_capability_info: CapabilityInfo):
    for request in mock_capability_info.session.query(CapabilityRequest).all():
        execution = mock_capability_info.create_execution(request)
        mock_capability_info.save_entity.assert_called_with(execution)


def test_lookup_capability(mock_capability_info: CapabilityInfo):
    for expected_capability in mock_capability_info.session.query(Capability).all():
        actual_capability = mock_capability_info.lookup_capability(expected_capability.name)
        assert expected_capability.__json__() == actual_capability.__json__()


def test_lookup_capability_request(
    mock_capability_info: CapabilityInfo, mock_capability_requests: List[CapabilityRequest]
):
    for expected_request in mock_capability_requests:
        actual_request = mock_capability_info.lookup_capability_request(expected_request.id)
        assert expected_request.__json__() == actual_request.__json__()


@given(
    st.sampled_from(SAMPLE_CAPABILITY_NAMES),
    st.lists(
        st.from_type(CapabilityRequest),
        min_size=1,
    ),
)
def test_lookup_active_capability_requests(
    mock_capability_info: CapabilityInfo, cap_name: str, list_of_request_json: List[Dict]
):
    """
    Given a capability name
    When the list of active requests is requested
    Then a list of capability requests with an active state (!= Complete) should be returned
    """
    # Insert requests into DB
    requests = [CapabilityRequest.from_json(blob) for blob in list_of_request_json]
    for request in requests:
        if not mock_capability_info.lookup_capability_request(request.id):
            mock_capability_info.session.add(request)
    mock_capability_info.session.commit()

    all_requests = mock_capability_info.session.query(CapabilityRequest).all()
    expected_actives = [r.__json__() for r in all_requests if r.state != "Complete" and r.capability_name == cap_name]

    active_requests_json = mock_capability_info.lookup_active_capability_requests(cap_name)
    assert active_requests_json["active_requests"] == expected_actives


@given(
    st.sampled_from(SAMPLE_CAPABILITY_NAMES),
    st.lists(
        st.from_type(CapabilityRequest),
        min_size=1,
    ),
)
def test_lookup_created_capability_requests(
    mock_capability_info: CapabilityInfo, cap_name: str, list_of_request_json: List[Dict]
):
    """
    Given a capability name
    When the list of active requests is requested
    Then a list of capability requests with an active state (!= Complete) should be returned
    """
    # Insert requests into DB
    requests = [CapabilityRequest.from_json(blob) for blob in list_of_request_json]
    for request in requests:
        if not mock_capability_info.lookup_capability_request(request.id):
            mock_capability_info.session.add(request)
    mock_capability_info.session.commit()

    all_requests = mock_capability_info.session.query(CapabilityRequest).all()
    expected_createds = [r.__json__() for r in all_requests if r.state == "Created" and r.capability_name == cap_name]

    active_requests_json = mock_capability_info.lookup_created_capability_requests(cap_name)
    assert active_requests_json["created_requests"] == expected_createds


@given(st.one_of(st.lists(st.text()), st.none()))
def test_lookup_current_version(
    mock_capability_info: CapabilityInfo,
    mock_capability_requests: List[CapabilityRequest],
    parameters: Optional[List[str]],
):
    """
    Given a list of versions for a request
    When the latest version is queried
    Then the version with the highest version number should be returned
    """
    for request in mock_capability_requests:
        mock_capability_info.create_capability_version(request.id, parameters)
        all_versions = (
            mock_capability_info.session.query(CapabilityVersion).filter_by(capability_request_id=request.id).all()
        )
        max_version_number = max(version.version_number for version in all_versions)
        assert mock_capability_info.lookup_current_version(request.id).version_number == max_version_number


@given(st.integers(min_value=0, max_value=SQLITE_MAX_INT))
@pytest.mark.usefixtures("mock_capability_execution")
def test_lookup_execution(mock_capability_info: CapabilityInfo, execution_id: int):
    """
    Given an execution ID
    When the database is queried for an execution with that ID
    Then an execution with that ID should be found if one exists
    """
    all_executions = mock_capability_info.session.query(CapabilityExecution).all()
    found_execution = mock_capability_info.lookup_execution(execution_id)

    for execution in all_executions:
        if execution_id == execution.id:
            assert found_execution.__json__() == execution.__json__()


@given(st.from_type(CapabilityRequest))
def test_save_entity(mock_capability_info: CapabilityInfo, request_to_save_json: CapabilityRequest):
    """
    Given a save-able entity
    When the entity is requested to be saved
    And the saving is performed
    Then it should be findable in the database
    """
    request_to_save = CapabilityRequest.from_json(request_to_save_json)
    if mock_capability_info.lookup_capability_request(request_to_save.id):
        # Remove existing entry to ensure a new entry is truly added
        mock_capability_info.session.query(CapabilityRequest).filter_by(id=request_to_save.id).delete()

    mock_capability_info.save_entity(request_to_save)
    assert mock_capability_info.session.query(CapabilityRequest).filter_by(id=request_to_save.id)


@given(st.from_type(CapabilityExecution), st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT))
def test_lookup_execution_by_wf_request_id(
    mock_capability_info: CapabilityInfo, execution_json: Dict, wf_request_id: int
):
    """
    Given a workflow request ID
    When the execution for the given ID is requested
    Then the corresponding execution should be returned
    Or None should be returned in the absence of a corresponding execution
    """
    # Remove existing entry for WF request so we start with a clean slate
    mock_capability_info.session.query(CapabilityExecution).filter_by(
        current_workflow_request_id=wf_request_id
    ).delete()

    execution = CapabilityExecution.from_json(execution_json)
    execution.current_workflow_request_id = wf_request_id

    if mock_capability_info.lookup_execution(execution.id):
        # Clear out execution with same ID as current, if one exists
        mock_capability_info.session.query(CapabilityExecution).filter_by(id=execution.id).delete()

    mock_capability_info.session.add(execution)

    assert (
        mock_capability_info.lookup_execution_by_workflow_request_id(wf_request_id).__json__() == execution.__json__()
    )


@given(
    st.lists(st.from_type(CapabilityRequest), min_size=3, max_size=3, unique_by=lambda req: req["id"]),
    st.sampled_from(SAMPLE_CAPABILITY_NAMES),
)
def test_requests_for_capability(mock_capability_info: CapabilityInfo, request_json_set: Set[Dict], cap_name: str):
    """
    Given a capability name
    When requests for the given capability are requested
    Then every capability request associated with the given capability name should be returned
    Or None should be returned in the absence of any corresponding requests
    """
    mock_capability_info.session.query(CapabilityRequest).filter_by(capability_name=cap_name).delete()
    requests = [CapabilityRequest.from_json(request_json) for request_json in request_json_set]

    for request in requests:
        request.capability_name = cap_name
        if mock_capability_info.lookup_capability_request(request.id):
            # Clear out request with same ID as current, if one exists
            mock_capability_info.session.query(CapabilityRequest).filter_by(id=request.id).delete()
        mock_capability_info.session.add(request)

    requests_for_cap = mock_capability_info.requests_for_capability(cap_name)

    assert all(expected_req.__json__() in [r.__json__() for r in requests_for_cap] for expected_req in requests)


@given(
    st.from_type(CapabilityExecution),
    st.from_type(CapabilityRequest),
    st.from_type(CapabilityVersion),
    st.from_type(CapabilityVersionFile),
)
def test_saving_various_entities(
    mock_capability_info: CapabilityInfo,
    execution_json: Dict,
    request_json: Dict,
    version_json: Dict,
    version_file_json: Dict,
):
    """
    Given a capability execution
    Or a capability request
    Or a capability version file
    When the entity is requested to be saved
    And it is saved
    Then it should be findable in the database
    """
    # Decode JSON into objects
    execution_to_save = CapabilityExecution.from_json(execution_json)
    request_to_save = CapabilityRequest.from_json(request_json)
    version = CapabilityVersion.from_json(version_json)
    version_file_to_save = CapabilityVersionFile.from_json(version_file_json)
    version_file_to_save.capability_request_id = version.capability_request_id
    version_file_to_save.version_number = version.version_number
    # Remove existing entries to ensure a new entry is truly added
    mock_capability_info.session.query(CapabilityExecution).filter_by(id=execution_to_save.id).delete()
    mock_capability_info.session.query(CapabilityRequest).filter_by(id=request_to_save.id).delete()
    mock_capability_info.session.query(CapabilityVersionFile).delete()
    # Save entries
    mock_capability_info.save_execution(execution_to_save)
    mock_capability_info.save_request(request_to_save)
    mock_capability_info.save_version_file(version, version_file_to_save.filename, version_file_to_save.content)
    # Ensure they can be found
    found_execution = mock_capability_info.session.query(CapabilityExecution).filter_by(id=execution_to_save.id).first()
    found_request = mock_capability_info.session.query(CapabilityRequest).filter_by(id=request_to_save.id).first()
    found_version_file = mock_capability_info.session.query(CapabilityVersionFile).first()

    assert execution_to_save.__json__() == found_execution.__json__()
    assert request_to_save.__json__() == found_request.__json__()
    assert version_file_to_save.__json__() == found_version_file.__json__()


@given(
    st.from_type(CapabilityVersion),
    st.lists(st.from_type(CapabilityVersionFile), min_size=3, max_size=3, unique_by=lambda file: file["filename"]),
)
def test_copy_previous_version_files(
    mock_capability_info: CapabilityInfo, version_json: Dict, list_of_version_file_json: List[Dict]
):
    """
    Given a capability version and a list of previous version files
    When the files are requested to be saved
    Then the files should be findable in the database
    And they should be associated with the given version
    """
    version = CapabilityVersion.from_json(version_json)
    version_files = [CapabilityVersionFile.from_json(blob) for blob in list_of_version_file_json]
    for file in version_files:
        file.capability_request_id = version.capability_request_id
        file.version_number = version.version_number

    # Clear out previous entities to start with a clean state
    mock_capability_info.session.query(CapabilityVersionFile).delete()
    # Save version files
    mock_capability_info.copy_previous_version_files(version, version_files)

    all_version_files = mock_capability_info.session.query(CapabilityVersionFile).all()
    assert all(
        expected_ver.__json__() == found_ver.__json__()
        for expected_ver, found_ver in zip(version_files, all_version_files)
    )


@pytest.mark.skip(reason="Not yet implemented")
def test_delete_request_versions():
    """
    Given a capability request ID
    When the versions of the given request are requested to be deleted
    Then the given request should have no versions associated with it after the deletion
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_delete_capability_request():
    """
    Given a capability request ID
    When the given request is requested to be deleted
    Then the given request should no longer be present in the database
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_can_delete_capability_request():
    """
    Given a capability request ID
    And the given request has no associated executions
    When the possibility of deleting the request is queried
    Then the answer should be True
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_cannot_delete_capability_request():
    """
    Given a capability request ID
    And the given request has associated executions
    When the possibility of deleting the request is queried
    Then the answer should be False
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_report():
    """
    Given a database state with capability entries
    When a queue report is requested for the database
    Then for each capability, the capability name
    And a count of its waiting executions ("Queued" state)
    And a count of its running executions ("Running" state)
    And the count of its max allowed jobs should be returned
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_all_waiting_executions():
    """
    Given a capability name
    And a limit that is less than the count of its waiting executions
    When the waiting executions are requested for the given capability
    Then a count of its waiting executions ("Queued" state) is returned
    """
    raise NotImplementedError


@pytest.mark.skip(reason="Not yet implemented")
def test_waiting_executions_limited():
    """
    Given a capability name
    And a limit that is greater than the count of its waiting executions
    When the waiting executions are requested for the given capability
    Then a count of the first *limit* waiting executions ("Queued" state) is returned
    """
    raise NotImplementedError
